# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2017-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

logger = logging.getLogger(__name__)

import os
import gzip
from importlib import import_module
import json
import collections

from aquaduct.utils.helpers import SmartRange, SmartRangeIncrement, SmartRangeEqual


################################################################################

class GlobalConfigStore(object):
    _cachedir = None
    _cachemem = False
    netcdf = False
    nc4 = False
    sandwich_import = False

    @property
    def cachedir(self):
        return self._cachedir

    @cachedir.setter
    def cachedir(self, value):
        assert not self.sandwich_import, "Cachedir property cannot be set after sandiwch import."
        self._cachedir = value

    @property
    def cachemem(self):
        return self._cachemem

    @cachemem.setter
    def cachemem(self, value):
        assert not self.sandwich_import, "Cachemem property cannot be set after sandiwch import."
        self._cachemem = value


GCS = GlobalConfigStore()


class CoordsRangeIndexCache(object):
    cache = {}

    def get_frc(self, number, rid):
        # wrapper for get ranges from frc
        if number not in self.cache:
            self.cache.update({number: {}})
            logger.debug("CRIC new number %d", number)
        if rid not in self.cache[number]:
            self.cache[number].update({rid: FramesRangeCollection()})
            logger.debug("CRIC new rid %d", rid)
        return self.cache[number][rid]

    def update_cric(self, cric):
        for number, rid_frc in cric.cache.iteritems():
            for rid, frc in rid_frc.iteritems():
                this_frc = self.get_frc(number, rid)
                for srange in frc.collection:
                    this_frc.append(srange)

    def setstate(self, state):

        for number, rid_frc in state:
            for rid, frc in rid_frc:
                this_frc = self.get_frc(number, rid)
                for srange in SmartRange(fast_minc_seq=frc).raw:
                    this_frc.append(srange)

    def reset(self):
        self.cache = {}

    def getstate(self):
        # this is saved as json so it constitutes format
        return ((int(k),
                 ((int(kk), map(int, SmartRange.raw2sequence(SmartRange(fast_raw=vv.collection).raw_increment))) for
                  kk, vv
                  in v.iteritems())) for k, v in self.cache.iteritems())


CRIC = CoordsRangeIndexCache()


################################################################################
# CIRC save in cache dir

def get_cric_reader(mode='r'):
    # returns file object
    if GCS.cachedir:
        try:
            data_file_name = GCS.cachedir + os.path.sep + 'cric.json'  # TODO: magic constant
            if mode == 'r' and not os.path.exists(data_file_name):
                return
            if mode == 'w' and not os.path.exists(GCS.cachedir):
                os.makedirs(GCS.cachedir)
            logger.debug("Preparing CRIC store with file %s", data_file_name)
            return gzip.open(data_file_name, mode=mode, compresslevel=9)
        except IOError:
            logger.warning("Unable to access CRIC data in cache dir [%s]." % GCS.cachedir)
            pass


def save_cric():
    vda = get_cric_reader(mode='w')
    if vda:
        logger.debug("Saving CRIC data.")
        json.dump(CRIC.getstate(), vda, cls=IterEncoder, indent=None, separators=(',', ':'))
        vda.close()


def load_cric():
    vda = get_cric_reader(mode='r')
    if vda:
        logger.debug("Loading CRIC data.")
        CRIC.reset()
        CRIC.setstate(json.load(vda))
        vda.close()
    else:
        CRIC.reset()


################################################################################
# FRC

class FramesRangeCollection(object):
    # currently it is assumed that samrt ranges increments only are possible
    def __init__(self):
        self.collection = []  # order on this list does matter!

    def append(self, srange):
        # TODO: remove it later
        if isinstance(srange, SmartRangeEqual):
            srange = SmartRangeIncrement(srange.element, srange.times)
        if not len(self.collection):
            self.collection.append(srange)
            logger.debug("FRC append first srange %s", str(srange))
            return
        # there are V cases:
        #           |------|            sr
        # 1 |---|                       it is before sr
        # 2      |-----|                it overlaps with sr but begins before
        # 3          |----|             it is contained in sr
        # 4              |----|         it overlaps with sr but ends after
        # 24     |------------|         it overlabs with sr in 2 and 4 way
        # 5                  |----|     it is after sr
        while srange is not None:
            for nr, sr in enumerate(self.collection):
                # sr
                if sr.overlaps_mutual(srange):  # or srange.overlaps(sr):
                    if sr.contains(srange):
                        # case 3
                        srange = None
                        break
                    if srange.first_element() < sr.first_element():
                        # case 2
                        self.collection.insert(nr, SmartRangeIncrement(srange.first_element(),
                                                                       sr.first_element() - srange.first_element()))
                        logger.debug("FRC case 2 insert srange %s at %d", str(
                            SmartRangeIncrement(srange.first_element(), sr.first_element() - srange.first_element())),
                                     nr)
                        if srange.last_element() > sr.last_element():
                            # case 24
                            srange = SmartRangeIncrement(sr.last_element() + 1,
                                                         srange.last_element() - sr.last_element())
                            break
                        else:
                            srange = None
                        break
                    if srange.last_element() > sr.last_element():
                        # case 4
                        srange = SmartRangeIncrement(sr.last_element() + 1, srange.last_element() - sr.last_element())
                        continue
                else:
                    if srange.last_element() < sr.first_element():
                        # case 1: insert it before sr
                        self.collection.insert(nr, srange)
                        logger.debug("FRC case 1 insert srange %s at %d", str(srange), nr)
                        srange = None
                        break
                    if srange.first_element() > sr.last_element():
                        # case 5: do nothing
                        continue
            # if something is left append it to the end
            if srange is not None and nr == len(self.collection) - 1:
                self.collection.append(srange)
                logger.debug("FRC append remaining srange %s", str(srange))
                srange = None

    def get_ranges(self, srange):
        # yield sranges from collection and appropriate ranges for these sranges
        # assumes append was already called? call it!
        # after it is called only case 3 or 4 is possible or no overlap at all
        self.append(srange)
        for sr in self.collection:
            if sr.overlaps(srange):
                # TODO: remove it later
                if isinstance(sr, SmartRangeEqual):
                    sr = SmartRangeIncrement(sr.element, sr.times)
                if sr.contains(srange):
                    # case 3
                    yield sr, xrange(srange.first_element() - sr.first_element(),
                                     srange.first_element() - sr.first_element() + len(srange))
                    srange = None
                    break
                # case 4
                yield sr, xrange(srange.first_element() - sr.first_element(),
                                 srange.first_element() - sr.first_element() +
                                 sr.last_element() - srange.first_element() + 1)
                srange = SmartRangeIncrement(sr.last_element() + 1,
                                             srange.last_element() - sr.last_element())


class IterEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, collections.Iterable):
            return list(o)
        return super(IterEncoder, self).default(o)


################################################################################


def get_object_name(something):
    name_ = something.__module__
    if hasattr(something, '__name__'):
        name_ += '.' + something.__name__
    elif hasattr(something, '__class__'):
        if hasattr(something.__class__, '__name__'):
            name_ += '.' + something.__class__.__name__
    return name_


def get_object_from_name(name):
    module_name = '.'.join(name.split('.')[:-1])
    object_name = name.split('.')[-1]
    module = import_module(module_name)
    return getattr(module, object_name)

################################################################################



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk,
#                          Magdalena Ługowska, Sandra Gołdowska <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

logger = logging.getLogger(__name__)
import numpy as np
from collections import namedtuple, OrderedDict
from scipy.spatial.distance import pdist, squareform
import copy
from itertools import izip_longest

from aquaduct.utils.helpers import is_iterable, listify, lind
from aquaduct.utils import clui
from aquaduct.utils.maths import make_default_array
from aquaduct.geom.convexhull import SciPyConvexHull


class ProtoInletTypeCodes(object):
    surface = 'surface'
    internal = 'internal'

    incoming = 'inin'
    outgoing = 'inout'


class InletTypeCodes(ProtoInletTypeCodes):
    # TODO: Rework this and make it coherent with barber create_spheres.
    all_surface = tuple(((ProtoInletTypeCodes.surface, itype) for itype in
                         (ProtoInletTypeCodes.incoming, ProtoInletTypeCodes.outgoing)))
    all_internal = tuple(((ProtoInletTypeCodes.internal, itype) for itype in
                          (ProtoInletTypeCodes.incoming, ProtoInletTypeCodes.outgoing)))
    all_incoming = tuple(((itype, ProtoInletTypeCodes.incoming) for itype in
                          (ProtoInletTypeCodes.surface, ProtoInletTypeCodes.internal)))
    all_outgoing = tuple(((itype, ProtoInletTypeCodes.outgoing) for itype in
                          (ProtoInletTypeCodes.surface, ProtoInletTypeCodes.internal)))

    surface_incoming = (ProtoInletTypeCodes.surface, ProtoInletTypeCodes.incoming)
    internal_incoming = (ProtoInletTypeCodes.internal, ProtoInletTypeCodes.incoming)
    internal_outgoing = (ProtoInletTypeCodes.internal, ProtoInletTypeCodes.outgoing)
    surface_outgoing = (ProtoInletTypeCodes.surface, ProtoInletTypeCodes.outgoing)


# clusers can be:
# None  - no such cluster
# nr - 0 means outliers
# because of passing paths one more type could be considered ???
# this might be also done in a different way

class InletClusterGenericType(object):
    def __init__(self, inp, out):
        super(InletClusterGenericType, self).__init__()
        self.clusters = [inp, out]

    @property
    def input(self):
        return self.clusters[0]

    @property
    def output(self):
        return self.clusters[1]
        # return self.clusters[-1]

    @staticmethod
    def cluster2str(cl):
        if cl is None:
            return 'N'
        return '%d' % int(cl)

    def __getitem__(self, item):
        return self.clusters[item]

    def __len__(self):
        return len(self.clusters)

    def __str__(self):
        return ':'.join(map(self.cluster2str, list(self)[::2] + list(self)[::-2]))

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self))

    def make_val(self, base):
        val = 0.
        for nr, e in enumerate(tuple(self)[::-1]):
            if e is None:
                e = float(base) ** (nr + 1)
            else:
                e += 1
            for dummy_iter in range(nr + 1):
                e /= float(base)
            val += e
        return val

    def __cmp__(self, other):

        def sort_for_cmp(a, b):

            # this function compares two elements and returns 0 for a == b, 1 for a > b , -1 for a < b

            if a == b:
                return 0
            elif a is None:
                if b > 0:
                    return 1
                elif b == 0:
                    return -1
            elif b is None:
                if a > 0:
                    return -1
                elif a == 0:
                    return 1
            elif a == 0:
                if b != 0:
                    return 1
            elif b == 0:
                if a != 0:
                    return -1
            elif a > b:
                return 1
            elif a < b:
                return -1

                # firstly only inputs are sorted, in case of self.input == other.input the outputs are sorted

        value = sort_for_cmp(self.input, other.input)
        if value == 0:
            return sort_for_cmp(self.output, other.output)
        else:
            return value

    def __hash__(self):
        return hash(str(self))


class InletClusterExtendedType(InletClusterGenericType):
    def __init__(self, surfin, interin, interout, surfout):
        super(InletClusterExtendedType, self).__init__(surfin, surfout)
        self.clusters.extend([interin, interout])

    @property
    def generic(self):
        return InletClusterGenericType(*self.clusters[:2])


# InletOLD = namedtuple('Inlet', 'coords type reference frame') # for backward compatibility

class Inlet(object):
    def __init__(self, coords=None, inlet_type=None, reference=None, frame=None):
        # no none is allowed
        assert None not in [inlet_type, reference, frame], "Wrong Inlet init."
        self.coords = make_default_array(coords)
        self.type = inlet_type
        self.reference = reference
        self.frame = frame


class Inlets(object):
    # class for list of inlets
    def __init__(self, spaths, center_of_system=None, onlytype=InletTypeCodes.all_surface, passing=False, pbar=None):

        self.center_of_system = center_of_system
        self.onlytype = onlytype
        self.inlets_list = []
        self.inlets_ids = []
        self.clusters = []
        self.number_of_clustered_inlets = None
        self.spheres = []
        self.passing = passing

        self.tree = clui.SimpleTree()

        for spath in spaths:
            self.extend_inlets(spath)
            if pbar: pbar.next()

    def __iter__(self):
        return (inlet for inlet in self.inlets_list)

    def add_leaf_wrapper(self, name=None, message=None, toleaf=None):
        if name == 0:
            self.tree.add_leaf(name='(0)', message=message, toleaf=toleaf)
        else:
            self.tree.add_leaf(name=name, message=message, toleaf=toleaf)

    def resize_leaf_0(self):
        if 0 in self.clusters_list:
            if 0 not in self.tree.leafs_names:
                self.tree.add_leaf(name=0, message='size: %d' % self.clusters.count(0))
            else:
                self.tree.add_message(toleaf=0, message='size: %d' % self.clusters.count(0), replace=True)

    def add_message_wrapper(self, message=None, toleaf=None):
        self.tree.add_message(message=message, toleaf=toleaf)

    def extend_inlets(self, spath, onlytype=None):
        if not self.passing and spath.is_passing():
            return

        if onlytype is None:
            onlytype = self.onlytype

        added_list = []
        nr = len(self.inlets_list)
        for inlet in spath.get_inlets():
            if onlytype is not None:
                if inlet.type not in onlytype:
                    continue
            # is ther already clustering info?
            if len(self.clusters):
                self.clusters.append(0)
            self.inlets_list.append(inlet)
            self.inlets_ids.append(nr)
            added_list.append(nr)
            nr += 1
        return added_list

    def add_cluster_annotations(self, clusters):
        # this replaces clusters!
        assert len(clusters) == len(self.inlets_list)
        self.clusters = clusters
        self.tree = clui.SimpleTree()

    def add_outliers_annotations(self, new_clusters):
        assert len(new_clusters) == len(self.inlets_list)
        # this is meant for situation when some points from clusters are changed to outliers
        # lets loop over current clusters list
        for cluster, cluster_size in zip(self.clusters_list, self.clusters_size):
            if cluster == 0: continue
            # check if cluster was changed!
            if cluster not in new_clusters:
                # it was completly removed!
                self.add_message_wrapper(message='outliers detection', toleaf=cluster)
                self.add_leaf_wrapper(name=0, toleaf=cluster, message='size: %d' % cluster_size)
            elif self.clusters.count(cluster) != new_clusters.count(cluster):
                # some points were shifted to outliers
                self.add_message_wrapper(message='outliers detection', toleaf=cluster)
                self.add_leaf_wrapper(name=0, toleaf=cluster,
                                      message='size: %d' % (cluster_size - new_clusters.count(cluster)))
                # we need new cluster!
                new_cluster = max(new_clusters) + 1
                for nr, dummy in enumerate(new_clusters):
                    if new_clusters[nr] == cluster:
                        new_clusters[nr] = new_cluster
                self.add_leaf_wrapper(name=new_cluster, toleaf=cluster,
                                      message='size: %d' % (new_clusters.count(new_cluster)))
        self.clusters = new_clusters
        self.resize_leaf_0()

    def add_spheres(self, spheres):
        assert len(spheres) == len(self.inlets_list)
        self.spheres = spheres

    def get_inlets_references(self):
        return [inl.reference for inl in self.inlets_list]

    # basic properites

    @property
    def size(self):
        return len(self.inlets_list)

    @property
    def coords(self):
        return [inlet.coords.tolist() for inlet in self.inlets_list]

    @property
    def types(self):
        return [inlet.type for inlet in self.inlets_list]

    @property
    def refs(self):
        return [inlet.reference for inlet in self.inlets_list]

    @property
    def refs_names(self):
        return [inlet.reference.name for inlet in self.inlets_list]

    def call_clustering_method(self, method, data, spheres=None):
        # this method runs clustering method against provided data
        # if center_of_system was set then use distance matrix...
        if self.center_of_system is not None:
            return method(np.array(data) - self.center_of_system, spheres=spheres)
        return method(np.array(data), spheres=spheres)

    def get_flat_tree(self, message=None):
        st = clui.SimpleTree(name='all', message='size: %d' % self.size)
        st.add_message(message=message)
        [st.add_leaf(name=leaf, message='size: %d' % csize) for leaf, csize in
         zip(self.clusters_list, self.clusters_size)]
        return st

    def perform_clustering(self, method):
        # this do clean clustering, all previous clusters are discarded
        # 0 means outliers
        self.add_cluster_annotations(self.call_clustering_method(method, self.coords, spheres=self.spheres))
        self.number_of_clustered_inlets = len(self.clusters)
        clui.message("New clusters created: %s" % (' '.join(map(str, sorted(set(self.clusters))))))
        # renumber clusters
        # self.renumber_clusters()
        # return clusters as simple tree
        self.tree = self.get_flat_tree(message=str(method))

    def perform_reclustering(self, method, skip_outliers=False, skip_size=None):
        # this do reclustering of all clusters, if no cluster exists perform_clustering is called
        if len(self.clusters) == 0:
            return self.perform_clustering(method)
        for cluster in self.clusters_list:
            if skip_outliers and cluster == 0:
                clui.message('Skipping outliers.')
                continue
            # check cluster size and skip if does not fit to skip_thershold function
            cluster_size = float(self.clusters.count(cluster)) / self.size
            if skip_size is not None:
                if skip_size(cluster_size):
                    clui.message('Cluster %d of size %0.3f skipped.' % (cluster, cluster_size))
                    continue
            clui.message('Cluster %d of size %0.3f submitted to reclustering.' % (cluster, cluster_size))
            self.recluster_cluster(method, cluster)
        # number of cluster
        self.number_of_clustered_inlets = len(self.clusters)
        # renumber clusters
        # self.renumber_clusters()
        # return clusters as simple tree

    # CLUSTER
    def recluster_cluster(self, method, cluster):
        if cluster in self.clusters_list:
            logger.debug('Reclustering %d cluster: initial number of clusters %d.' % (cluster, len(self.clusters_list)))
            reclust = self.call_clustering_method(method, self.lim2clusters(cluster).coords,
                                                  spheres=self.lim2clusters(cluster).spheres)
            if len(set(reclust)) <= 1:
                clui.message('No new clusters found.')
            else:
                # how many clusters? what aboout outliers?
                n_reclust = len(set(reclust))
                out_reclust = 0 in reclust
                clui.message('Cluster %d was split into %d clusters.' % (cluster, n_reclust))
                if out_reclust:
                    clui.message('Outliers were detected and will have annotation of the old cluster %d.' % cluster)
                else:
                    clui.message('No outliers were detected, the old cluster %d will be removed.' % cluster)
                # change numbers of reclust
                max_cluster = max(self.clusters_list)
                for nr, r in enumerate(reclust):
                    if r != 0:
                        reclust[nr] = r + max_cluster
                if cluster != 0:
                    self.add_message_wrapper(message=str(method), toleaf=cluster)
                [self.add_leaf_wrapper(name=leaf, toleaf=cluster, message=['size: %d' % reclust.count(leaf)]) for leaf
                 in sorted(list(set(reclust)))]
                if cluster == 0:
                    self.add_message_wrapper(message=['[RE]', str(method)], toleaf=cluster)
                if out_reclust:
                    clui.message('The old cluster %d will be split into new clusters: %s' % (
                        cluster, (' '.join(map(str, sorted(set(reclust))[1:])))))
                else:
                    clui.message('The old cluster %d will be split into new clusters: %s' % (
                        cluster, (' '.join(map(str, sorted(set(reclust)))))))
                # add new clusters
                nrr = 0
                for nr, c in enumerate(self.clusters):
                    if c == cluster:
                        self.clusters[nr] = reclust[nrr]
                        nrr += 1
            logger.debug('Reclustering %d cluster: final number of clusters %d.' % (cluster, len(self.clusters_list)))
        # number of cluster
        self.number_of_clustered_inlets = len(self.clusters)
        if cluster != 0:
            self.resize_leaf_0()

    def recluster_outliers(self, method):
        self.recluster_cluster(method, 0)
        # renumber clusters
        # self.renumber_clusters()

    def remove_inlets(self, cluster):
        # removes inlets completly - clusters and inlets
        n = self.clusters.count(cluster)
        to_remove = []
        for nr, c in enumerate(self.clusters):
            if c == cluster:
                to_remove.append(nr)
        for tr in to_remove[::-1]:
            self.clusters.pop(tr)
            self.inlets_list.pop(tr)
            self.inlets_ids.pop(tr)
        self.add_message_wrapper(message='cluster %d removed' % cluster, toleaf=cluster)

    def small_clusters_to_outliers(self, maxsize):
        new_out = 0
        for c in self.clusters_list:
            if c == 0:
                continue
            if self.clusters.count(c) <= maxsize:
                self.add_leaf_wrapper(name=0, toleaf=c, message='size: %d' % self.clusters.count(c))
                self.add_message_wrapper(message='|%d| to outliers' % maxsize, toleaf=c)
                for nr, cc in enumerate(self.clusters):
                    if cc == c:
                        self.clusters[nr] = 0
                        new_out += 1
                        # renumber clusters
                        # self.renumber_clusters()
        if new_out:
            if 0 not in self.tree.leafs_names:
                self.tree.add_leaf(name=0)
            self.add_message_wrapper(message=['|%d| to outliers' % maxsize, 'new size: %d' % self.clusters.count(0)],
                                     toleaf=0)
            # self.resize_leaf_0()

    def renumber_clusters(self):
        old_clusters = list(self.clusters)
        if 0 in self.clusters_list:
            new_numbers = range(len(self.clusters_list))
        else:
            new_numbers = range(1, len(self.clusters_list) + 1)
        old_numbers = self.clusters_list
        if old_numbers != new_numbers:
            for nr, c in enumerate(self.clusters):
                self.clusters[nr] = new_numbers[old_numbers.index(c)]
        self.sort_clusters()
        # make old new dict
        ond = []
        for o, n in zip(old_clusters, self.clusters):
            if o not in ond:
                ond.append(o)
                self.add_message_wrapper(toleaf=o, message='new nr: %d' % n)

    def sort_clusters(self):
        old_numbers = self.clusters_list
        old_sizes = self.clusters_size
        # now sort according to sizes but put 0, if present, at the begining
        if 0 in old_numbers:
            zero_size = old_sizes.pop(old_numbers.index(0))
            zero_number = old_numbers.pop(old_numbers.index(0))  # which is zero!
            new_numbers = [zero_number] + [old_numbers[i] for i in np.argsort(old_sizes).tolist()[::-1]]
            new_sizes = [zero_size] + [old_sizes[i] for i in np.argsort(old_sizes).tolist()[::-1]]
            old_numbers = self.clusters_list
            # old_sizes = self.clusters_size
        else:
            new_numbers = [old_numbers[i] for i in np.argsort(old_sizes).tolist()[::-1]]
            new_sizes = [old_sizes[i] for i in np.argsort(old_sizes).tolist()[::-1]]
        # trans_dict = {n: o for o, n in zip(old_numbers, new_numbers)}
        trans_dict = dict((n, o) for o, n in
                          zip(old_numbers, new_numbers))  # more universal as dict comprehension may not work in <2.7
        new_clusters = []
        for c in self.clusters:
            new_clusters.append(trans_dict[c])
        self.clusters = new_clusters
        assert self.clusters_size == new_sizes

    @property
    def clusters_list(self):
        return sorted(list(set(self.clusters)))

    @property
    @listify
    def clusters_centers(self):
        for c in self.clusters_list:
            yield np.mean(self.lim2clusters(c).coords, 0)

    @property
    def clusters_size(self):
        return map(self.clusters.count, self.clusters_list)

    def join_clusters(self, clusters2join):
        new_cluster = max(self.clusters_list) + 1
        for nr, c in enumerate(self.clusters):
            if c in clusters2join:
                self.clusters[nr] = new_cluster
        for c in clusters2join:
            self.add_message_wrapper(toleaf=c, message='joined into: %d' % new_cluster)
        nc_size = self.clusters.count(new_cluster)
        self.add_leaf_wrapper(name=new_cluster, message='size: %d' % nc_size, toleaf=None)
        self.add_message_wrapper(message='made of %s' % ('+'.join(map(str, clusters2join))), toleaf=new_cluster)

    @property
    @listify
    def clusters_std(self):
        for c, s in zip(self.clusters_list, self.clusters_size):
            if s == 1:
                yield 0.
            elif s == 2:
                yield pdist(self.lim2clusters(c).coords, 'euclidean')
            elif s > 2:
                yield np.std(pdist(self.lim2clusters(c).coords, 'euclidean'))

    @listify
    def spaths2ctypes(self, spaths):
        # surfin interin interout surfout
        for sp in spaths:
            yield self.spath2ctype(sp)

    def spath2ctype(self, sp):
        ctypes = self.lim2spaths(sp).types
        clusters = self.lim2spaths(sp).clusters
        surfin, interin, interout, surfout = None, None, None, None
        for ct, cl in zip(ctypes, clusters):
            if ct in InletTypeCodes.all_surface and ct in InletTypeCodes.all_incoming:
                surfin = cl
            if ct in InletTypeCodes.all_internal and ct in InletTypeCodes.all_incoming:
                interin = cl
            if ct in InletTypeCodes.all_internal and ct in InletTypeCodes.all_outgoing:
                interout = cl
            if ct in InletTypeCodes.all_surface and ct in InletTypeCodes.all_outgoing:
                surfout = cl
        return InletClusterExtendedType(surfin, interin, interout, surfout)

    def lim_to(self, what, towhat):
        if not is_iterable(towhat):
            towhat = [towhat]
        new_inlets = self.__class__([], onlytype=self.onlytype,
                                    center_of_system=self.center_of_system,
                                    passing=self.passing,
                                    pbar=None)  # do not propagate pabr
        new_inlets.number_of_clustered_inlets = self.number_of_clustered_inlets

        for inlet, ids, cluster, sphere, w in izip_longest(self.inlets_list, self.inlets_ids, self.clusters,
                                                           self.spheres, what):
            if w in towhat:
                new_inlets.inlets_list.append(inlet)
                new_inlets.inlets_ids.append(ids)
                new_inlets.clusters.append(cluster)
                if len(self.spheres) and (sphere or sphere == 0):
                    new_inlets.spheres.append(sphere)

        return new_inlets

    def lim2spaths(self, spaths):
        if not is_iterable(spaths):
            spaths = [spaths]
        return self.lim_to(self.refs, [sp.id for sp in spaths])

    def lim2rnames(self, rnames):
        if not is_iterable(rnames):
            rnames = [rnames]
        return self.lim_to(self.refs_names, rnames)

    def lim2types(self, types):
        return self.lim_to(self.types, types)

    def lim2clusters(self, clusters):
        return self.lim_to(self.clusters, clusters)

    @listify
    def limspaths2(self, spaths):
        if not is_iterable(spaths):
            spaths = [spaths]
        refs = set(self.refs)
        for sp in spaths:
            if sp.id in refs:
                yield sp

    def get_chull(self):
        return SciPyConvexHull(self.coords)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

logger = logging.getLogger(__name__)

from itertools import izip
from collections import OrderedDict
from aquaduct.utils import clui
import numpy as np
from scipy.spatial.distance import cdist
from aquaduct.geom import traces
from aquaduct.traj.inlets import Inlet, InletTypeCodes
from aquaduct.utils.helpers import is_number, lind, glind, \
    SmartRange  # smart ranges are required here to provide backward compatibility with v0.3
from aquaduct.utils.sets import intersection, glue, left, right
from aquaduct.utils.helpers import tupleify, listify, arrayify1, list_blocks_to_slices
from aquaduct.utils.maths import make_default_array
from aquaduct.traj.sandwich import Reader, SingleResidueSelection
from array import array

from itertools import chain, imap


class PathTypesCodes(object):
    __slots__ = ()  # "path_in_code path_object_code path_out_code path_walk_code".split()
    # path_in_code = 'i'
    # path_object_code = 'c'
    # path_out_code = 'o'
    # path_walk_code = 'w'
    path_in_code = 1
    path_object_code = 2
    path_out_code = 3
    path_walk_code = 0


class GenericPathTypeCodes(object):
    __slots__ = ()  # "object_name scope_name out_name".split()
    # object_name = 'c'
    # scope_name = 's'
    # out_name = 'n'
    object_name = 8
    scope_name = 9
    out_name = 10


class GenericPaths(GenericPathTypeCodes):
    # object to store paths... is it required?
    __slots__ = 'id single_res_selection name _types _frames max_possible_frame min_possible_frame'.split()

    def __init__(self, id_of_res, name_of_res=None,
                 min_pf=None, max_pf=None):

        super(GenericPaths, self).__init__()

        # id is any type of object; it is used as identifier
        # single_res_selection is object which have coords method that accepts frames and returns coordinates

        self.id = id_of_res
        self.single_res_selection = SingleResidueSelection(self.id)

        if name_of_res is not None:
            self.name = name_of_res
        else:
            self.name = 'UNK'  # FIXME: magic constant
        assert isinstance(self.name, str)
        assert len(self.name) == 3
        self._types = SmartRange()
        # self._frames = SmartRange()
        self._frames = array('i')

        # following is required to correct in and out paths that begin or end in scope and
        # begin or end at the very begining of MD or at very end of MD

        self.max_possible_frame = max_pf
        self.min_possible_frame = min_pf

    def update_types_frames(self, types, frames):
        if isinstance(types, SmartRange) and isinstance(frames, SmartRange):
            self._types = types
            self._frames = array('i', list(frames.get()))
        elif isinstance(types, SmartRange) and isinstance(frames, array):
            self._types = types
            self._frames = frames
        # make it from "any" iterable
        else:
            self._types = SmartRange(types)
            self._frames = array('i', frames)
        assert len(self._frames) == len(set(self._frames)), (len(self._frames), len(set(self._frames)))

    def __getstate__(self):
        return self.id, self.name, self._types, self._frames, self.max_possible_frame, self.min_possible_frame

    def __setstate__(self, state):
        # FIXME: tmp solution
        if isinstance(state, dict):
            self.id = state['id']
            self.name = state['name']
            self._types = state['_GenericPaths_types']
            self._frames = state['_GenericPaths_frames']
            self.max_possible_frame = state['max_possible_frame']
            self.min_possible_frame = state['min_possible_frame']
        else:
            self.id, self.name, self._types, self._frames, self.max_possible_frame, self.min_possible_frame = state
        self.single_res_selection = SingleResidueSelection(self.id)

    # info methods
    @property
    def types(self):
        return list(self._types.get())

    @property
    def types_promise(self):
        return self._types.get()

    @property
    def frames_of_object(self):
        def get_foo():
            frame = self._frames.first_element()
            for sr in self._types.raw:
                if sr.element == self.object_name:
                    yield xrange(frame, frame + sr.times)
                frame += sr.times

        return chain(*get_foo())

    @property
    def frames_of_scope(self):
        def get_fos():
            frame = self._frames.first_element()
            for sr in self._types.raw:
                if sr.element == self.scope_name:
                    yield xrange(frame, frame + sr.times)
                frame += sr.times

        return chain(*get_fos())

    @property
    def frames(self):
        # return list(self._frames.get())
        return list(self._frames)

    @property
    def _frames_sr(self):
        return SmartRange(fast_array=self._frames)

    # @property
    # def frames_promise(self):
    #    return self._frames.get()

    def discard_singletons(self, singl=1, skiptype=GenericPathTypeCodes.object_name):
        # singl is chunk size to discard
        types = self.types
        new_types = []
        new_frames = []
        seek = 0
        # loop over frames chunks:
        for chunksr in self._frames_sr.raw:
            if len(chunksr) <= singl:
                if skiptype not in types[seek:seek + len(chunksr)]:
                    continue

            new_frames.append(chunksr)
            new_types.extend(types[seek:seek + len(chunksr)])
            seek += len(chunksr)

        self._types = SmartRange(new_types)
        self._frames = array('i', chain(*(ch.get() for ch in new_frames)))

    @property
    def coords(self):
        return self.single_res_selection.coords(self._frames_sr)

    @property
    def max_frame(self):
        return max(self._frames)

    @property
    def min_frame(self):
        return min(self._frames)

    # add methods
    def add_foos(self, foo, fos):
        foo_ = None
        fos_ = None
        while len(foo) + len(fos):
            if len(foo) and foo_ is None:
                foo_ = foo.pop(0)
            if len(fos) and fos_ is None:
                fos_ = fos.pop(0)
            if foo_ is None:
                self.add_scope(fos_)
                fos_ = None
            elif fos_ is None:
                self.add_object(foo_)
                foo_ = None
            elif foo_ > fos_:
                self.add_scope(fos_)
                fos_ = None
            elif fos_ > foo_:
                self.add_object(foo_)
                foo_ = None

    def add_012(self, os_in_frames, reset=False):
        if reset:
            self._types = SmartRange()
            self._frames = array('i')
        for frame, os_type in enumerate(os_in_frames):
            if os_type == 2:
                self.add_object(frame)
            elif os_type == 1:
                self.add_scope(frame)

    def add_object(self, frame):
        self.add_type(frame, self.object_name)

    def add_scope(self, frame):
        self.add_type(frame, self.scope_name)

    def add_type(self, frame, ftype):
        assert frame not in self._frames
        self._types.append(ftype)
        self._frames.append(frame)

    def add_frames_types(self, frames, types):
        for f, t in izip(frames, types):
            self._types.append(t)
            self._frames.append(f)

    '''
    def _split_path_by_edges_(self,path):
        # input: path tuple
        # output: path_cont
        edges = self.single_res_selection.get_edges()
        path_cont = []
        map(path_cont.extend, path)
        if edges:
            for e in edges:
                if e in path_cont:
                    path_cont_2y = path_cont[:path_cont.index(e) + 1]
                    path_cont = path_cont[path_cont.index(e) + 1:]
                    yield path_cont_2y
        if path_cont:
            yield path_cont
    '''

    '''
    def _split_path_by_types_(self,path):
        # input: path_cont
        # output: path tuple
        types = self.get_path_cont_types(path)
        # now, split it into single path or passing
        if self.object_name in types:
            # in
            bp = types.index(self.object_name)
            in_p = path[:bp]
            path = path[bp:]
            types = types[bp:]
            # obj
            types = types[::-1]
            path = path[::-1]
            ep = types.index(self.object_name)
            out_p = path[:ep]
            path = path[ep:]
            out_p = out_p[::-1]
            path = path[::-1]
            return in_p,path,out_p
        else:
            return (path,[],[]) # passing
    '''

    '''
    def _consider_edges(self,path,passing=None):
        for path_cont in self._split_path_by_edges_(path):
            spath = self._split_path_by_types_(path_cont)
            if len(spath[1])>0 or passing:
                yield spath
    '''

    def _gpt(self):
        # get, I'm just passing through
        n = len(self._frames)
        types = self.types
        begin = 0
        for block in self._frames_sr.raw:
            end = begin + block.times
            # get types of this block
            block_frames = list(block.get())
            block_types = types[begin:end]
            begin = end
            # now iterate over block_types in a search of out_name
            if self.out_name in block_types:
                while self.out_name in block_types:
                    to_yield = block_frames[:block_types.index(self.out_name)]
                    to_yield_types = block_types[:block_types.index(self.out_name)]
                    if len(to_yield) > 0:
                        if self.object_name not in to_yield_types:
                            yield to_yield
                    block_types = block_types[block_types.index(self.out_name):]
                    block_frames = block_frames[block_types.index(self.out_name):]
            if len(block_frames) > 0:
                if len(block_frames) > 0:
                    if self.object_name not in block_types:
                        yield block_frames

    def _gpo(self, frames_sr):
        n = len(frames_sr)
        types = self.types
        begin = 0
        for block in frames_sr.raw:
            end = begin + block.times
            # get types of this block
            block_frames = list(block.get())
            block_types = types[begin:end]
            begin = end
            # now iterate over block_types in a search of out_name
            if self.out_name in block_types:
                while self.out_name in block_types:
                    to_yield = block_frames[:block_types.index(self.out_name)]
                    to_yield_types = block_types[:block_types.index(self.out_name)]
                    if len(to_yield) > 0:
                        while to_yield_types[0] != self.object_name:
                            to_yield.pop(0)
                            to_yield_types.pop(0)
                            if len(to_yield) == 0:
                                break
                        if len(to_yield) > 0:
                            yield to_yield
                    block_types = block_types[block_types.index(self.out_name):]
                    block_frames = block_frames[block_types.index(self.out_name):]
            if len(block_frames) > 0:
                while block_types[0] != self.object_name:
                    block_frames.pop(0)
                    block_types.pop(0)
                    if len(block_frames) == 0:
                        break
                if len(block_frames) > 0:
                    yield block_frames

    def _gpi(self, frames_sr):
        n = len(frames_sr)
        types = self.types
        begin = 0
        for block in frames_sr.raw:
            end = begin + block.times
            # get types of this block
            block_frames = list(block.get())
            block_types = types[begin:end]
            begin = end
            # now iterate over block_types in a search of out_name
            if self.out_name in block_types:
                while self.out_name in block_types:
                    to_yield = block_frames[:block_types.index(self.out_name)]
                    to_yield_types = block_types[:block_types.index(self.out_name)]
                    if len(to_yield) > 0:
                        while to_yield_types[-1] != self.object_name:
                            to_yield.pop(-1)
                            to_yield_types.pop(-1)
                            if len(to_yield) == 0:
                                break
                        if len(to_yield) > 0:
                            yield to_yield
                    block_types = block_types[block_types.index(self.out_name):]
                    block_frames = block_frames[block_types.index(self.out_name):]
            if len(block_frames) > 0:
                while block_types[-1] != self.object_name:
                    block_frames.pop(-1)
                    block_types.pop(-1)
                    if len(block_frames) == 0:
                        break
                if len(block_frames) > 0:
                    yield block_frames

    def find_paths(self, fullonly=False):
        # this looks for normal, ie containing 'core' paths
        frames_sr = self._frames_sr
        paths_out = list(self._gpo(frames_sr))
        paths_in = list(self._gpi(frames_sr))
        del frames_sr

        for path_in in paths_in:
            path_out = paths_out[0]
            path_glue = glue(path_in, path_out)
            if len(path_glue) > 0:
                path_out = paths_out.pop(0)
                path_core = intersection(path_in, path_out)
                path_in = left(path_in, path_out)
                path_out = right(path_core, path_out)
            else:
                path_core = []
                path_out = []
            if len(path_in) == 0 or len(path_core) == 0 or len(path_out) == 0:
                if fullonly:
                    continue
            # TODO: make separate method/function for that
            # correct for max/min possible frame
            if self.max_possible_frame is not None:
                if len(path_out) > 0:
                    if path_out[-1] >= self.max_possible_frame:
                        path_core += path_out
                        path_out = []
            if self.min_possible_frame is not None:
                if len(path_in) > 0:
                    if path_in[0] <= self.min_possible_frame:
                        path_core = path_in + path_core
                        path_in = []
            yield path_in, path_core, path_out
        if not fullonly:
            path_in = []
            path_core = []
            for path_out in paths_out:
                # correct for max/min possible frame
                if self.max_possible_frame is not None:
                    if len(path_out) > 0:
                        if path_out[-1] >= self.max_possible_frame:
                            path_core += path_out
                            path_out = []
                yield path_in, path_core, path_out

    def find_paths_types(self, fullonly=False, passing=None):
        # both loops below yields correct paths
        # if waterfall let's split each of yielded paths again
        edges = self.single_res_selection.get_edges()
        for path in self.find_paths(fullonly=fullonly):
            '''
            for path_sbe in self._consider_edges(path,passing):
                types = self.get_single_path_types(path_sbe)
                # yield path, coords, types
                yield path_sbe, types
            '''
            types = self.get_single_path_types(path)
            # yield path, coords, types
            yield path, types
        if passing:
            for path in self._gpt():
                '''
                for path_sbe in self._consider_edges((path, [], []), passing):
                    types = self.get_single_path_types(path_sbe)
                    # yield path, coords, types
                    yield path_sbe, types
                '''
                types = self.get_single_path_types((path, [], []))
                # yield path, coords, types
                yield (path, [], []), types

    def get_path_cont_types(self, path_cont):

        frames = self.frames
        types = self.types

        def get_be(p_):
            # if len(p_) == 1:
            # quit    return 0,1
            return frames.index(p_[0]), frames.index(p_[-1]) + 1

        if len(frames) == len(types):
            # not full trajectory
            if len(path_cont) > 0:
                b, e = get_be(path_cont)
                return types[b:e]
        else:
            # full trajectory
            if len(path_cont) > 0:
                b, e = path_cont[0], path_cont[-1] + 1
                return types[b:e]
        return []

    def get_single_path_types(self, spath):
        # returns typess for single path
        # single path comprises of in,scope,out parts

        # TODO: join it with get_single_path_coords

        p_in, p_object, p_out = spath

        frames = self.frames
        types = self.types

        def get_be(p_):
            # if len(p_) == 1:
            # quit    return 0,1
            return frames.index(p_[0]), frames.index(p_[-1]) + 1

        in_t = []
        object_t = []
        out_t = []
        in_c = []
        object_c = []
        out_c = []

        if len(frames) == len(types):
            # not full trajectory
            # p_in
            if len(p_in) > 0:
                b, e = get_be(p_in)
                in_t = types[b:e]
                # in_c = self.coords[b:e]
            # p_object
            if len(p_object) > 0:
                b, e = get_be(p_object)
                object_t = types[b:e]
                # object_c = self.coords[b:e]
            # p_out
            if len(p_out) > 0:
                b, e = get_be(p_out)
                out_t = types[b:e]
                # out_c = self.coords[b:e]
        else:
            # full trajectory
            # p_in
            if len(p_in) > 0:
                b, e = p_in[0], p_in[-1] + 1
                in_t = types[b:e]
                # in_c = self.coords[b:e]
            # p_object
            if len(p_object) > 0:
                b, e = p_object[0], p_object[-1] + 1
                object_t = types[b:e]
                # object_c = self.coords[b:e]
            # p_out
            if len(p_out) > 0:
                b, e = p_out[0], p_out[-1] + 1
                out_t = types[b:e]
                # out_c = self.coords[b:e]

        # return (in_c, object_c, out_c), (in_t, object_t, out_t)
        return in_t, object_t, out_t

    def barber_with_spheres(self, spheres):
        # calculate big distance matrix
        # distances = cdist(self.coords, [s.center for s in spheres],metric='euclidean')
        # compare with radii
        # tokeep = distances > np.matrix([[s.radius for s in spheres]])
        if len(spheres):
            tokeep = np.argwhere(
                (cdist(np.array(list(self.coords)), [s.center for s in spheres], metric='euclidean') > np.matrix(
                    [[s.radius for s in spheres]])).all(1).A1).flatten().tolist()
            # tokeep = np.argwhere(tokeep).flatten().tolist()
            self._types = SmartRange(lind(self.types, tokeep))
            self._frames = SmartRange(lind(self.frames, tokeep))


# SinglePathID = namedtuple('SinglePathID', 'id nr')
class SinglePathID(object):
    __slots__ = 'id nr name'.split()

    def __init__(self, path_id=None, nr=None, name=None):
        assert path_id is not None, "path_id connot be None."
        self.id = path_id
        assert nr is not None, "nr connot be None."
        self.nr = nr
        assert name is not None, "name connot be None."
        self.name = name

    def __getstate__(self):
        return self.id, self.nr, self.name

    def __setstate__(self, state):
        self.id, self.nr, self.name = state

    def __str__(self):
        # by default name is not returned
        return '%d:%d:%d' % (self.id + (self.nr,))

    def __eq__(self, other):
        if isinstance(other, SinglePathID):
            return self.id == other.id and self.nr == other.nr and self.name == other.name
        return False


@listify
def yield_single_paths(gps, fullonly=None, progress=None, passing=None):
    # iterates over gps - list of GenericPaths objects and transforms them in to SinglePath objects
    nr_dict = {}
    for nr, gp in enumerate(gps):
        path_id = gp.id
        path_name = gp.name
        with clui.tictoc('Processing path %d:%d' % path_id):
            for paths, types in gp.find_paths_types(fullonly=fullonly, passing=passing):
                # for paths, coords, types in gp.find_paths_types(fullonly=fullonly):
                if path_id in nr_dict:
                    nr_dict.update({path_id: (nr_dict[path_id] + 1)})
                else:
                    nr_dict.update({path_id: 0})

                pid = SinglePathID(path_id=path_id, nr=nr_dict[path_id], name=path_name)

                if len(paths[1]) > 0:
                    # if everything is OK
                    sp = SinglePath(pid, paths, types)
                else:
                    # this is passing through
                    if passing:
                        # import ipdb as pdb; pdb.set_trace()
                        sp = PassingPath(pid, paths, types)
                        sp.has_in = sp.paths_first_in > gp.min_possible_frame
                        sp.has_out = sp.paths_last_out < gp.max_possible_frame
                    else:
                        continue
                if progress:
                    yield sp, nr
                else:
                    yield sp


def correct_spaths_ids(spaths, pbar):
    seen = {}
    for sp in spaths:
        if sp.id.id not in seen:
            seen.update({sp.id.id: 0})
        else:
            seen[sp.id.id] += 1
        sp.id.nr = seen[sp.id.id]
        pbar.next()
        # print str(sp.id)


def yield_generic_paths(spaths, progress=None):
    rid_seen = OrderedDict()
    number_of_frames = Reader.number_of_frames(onelayer=True) - 1
    t_object = GenericPathTypeCodes.object_name
    for sp in spaths:
        current_rid = sp.id.id
        if current_rid not in rid_seen:
            rid_seen.update(
                {current_rid: GenericPaths(current_rid, name_of_res=sp.id.name, min_pf=0, max_pf=number_of_frames)})
        # TODO: following loop is not an optimal solution, it is better to add types and frames in one call
        for t, f in zip(sp.gtypes_cont, sp.paths_cont):
            if t == t_object:
                rid_seen[current_rid].add_object(f)
            else:
                rid_seen[current_rid].add_scope(f)
        if progress:
            progress.next()
    # because paths stores now frames as array and produces smartranges on demand with fast_array option
    # it is required to keep frames (and types) in order, otherwise smartranges are wrong
    for p in rid_seen.itervalues():
        new_order = np.argsort(p.frames)
        p.update_types_frames(glind(p.types, new_order), glind(p.frames, new_order))
        progress.next()

    return rid_seen.values()


class MacroMolPath(PathTypesCodes, InletTypeCodes):
    # special class
    # represents one path

    __slots__ = "id _path_in _path_object _path_out _types_in _types_object _types_out _object_len single_res_selection".split()

    empty_coords = make_default_array(np.zeros((0, 3)))

    def __init__(self, path_id, paths, types):

        super(MacroMolPath, self).__init__()

        self.id = path_id
        self.single_res_selection = SingleResidueSelection(self.id.id)
        # for paths use SmartRanges and then provide methods to read them
        self._path_in, self._path_object, self._path_out = map(SmartRange, paths)
        # similarly, do it with types
        # self.path_in, self.path_object, self.path_out = paths
        self._types_in, self._types_object, self._types_out = map(SmartRange, types)
        # self.types_in, self.types_object, self.types_out = types

        # self.coords_in, self.coords_object, self.coords_out = map(make_default_array, coords)

        # self.smooth_coords_in, self.smooth_coords_object, self.smooth_coords_out = None, None, None
        # self.smooth_method = None

        # return np.vstack([c for c in self._coords if len(c) > 0])

        self._object_len = None

    def __getstate__(self):
        return self.id, self._path_in, self._path_object, self._path_out, self._types_in, self._types_object, self._types_out, self._object_len

    def __setstate__(self, state):
        self.id, self._path_in, self._path_object, self._path_out, self._types_in, self._types_object, self._types_out, self._object_len = state
        self.single_res_selection = SingleResidueSelection(self.id.id)

    def add_paths4(self, path_in, path_object, path_object_strict, path_out):
        # init empty path
        self.__path_in, self.__path_object, self.__path_out = map(SmartRange, (path_in, path_object, path_out))
        self.__types_in = SmartRange([GenericPathTypeCodes.scope_name] * len(path_in))
        self.__types_out = SmartRange([GenericPathTypeCodes.scope_name] * len(path_out))

        tobj = np.array([GenericPathTypeCodes.scope_name] * len(path_object))
        tobj[[nr for nr, f in enumerate(path_object) if f in path_object_strict]] = GenericPathTypeCodes.object_name
        self.__types_object = SmartRange(tobj.tolist())

    def _object_len_calculate(self):
        for nr, real_coords in enumerate(traces.midpoints(self.coords)):
            if nr != 1:
                continue
            if len(real_coords) <= 1:
                return 0.
            return float(sum(traces.diff(real_coords)))

    @property
    def object_len(self):
        if self._object_len is None:
            self._object_len = self._object_len_calculate()
        return self._object_len

    def is_single(self):
        raise NotImplementedError("Implementation missing.")

    def is_passing(self):
        raise NotImplementedError("Implementation missing.")

    def is_frame_in(self, frame):
        return self._path_in.isin(frame)

    def is_frame_object(self, frame):
        return self._path_object.isin(frame)

    def is_frame_out(self, frame):
        return self._path_out.isin(frame)

    def is_frame_walk(self, frame):
        return self.is_frame_in(frame) or self.is_frame_object(frame) or self.is_frame_out(frame)

    @property
    def path_in(self):
        return list(self._path_in.get())

    @property
    def path_object(self):
        return list(self._path_object.get())

    @property
    def path_object_strict_len(self):
        # number of frames strictly in object
        return self.etypes[1].count(self.etypes[1][0])

    @property
    def path_out(self):
        return list(self._path_out.get())

    @property
    def types_in(self):
        return list(self._types_in.get())

    @property
    def types_object(self):
        return list(self._types_object.get())

    @property
    def types_out(self):
        return list(self._types_out.get())

    # ---------------------------------------------------------------------------------------------------------------- #

    @property
    def coords_first_in(self):
        if len(self._path_in) > 0:
            return self.coords_in[0]
            # return self.single_res_selection.coords([self._path_in.first_element()])

    @property
    def paths_first_in(self):
        if len(self._path_in) > 0:
            return self.path_in[0]

    @property
    def coords_last_out(self):
        if len(self._path_out) > 0:
            return self.coords_out[-1]
            # return self.single_res_selection.coords([self._path_out.last_element()])

    @property
    def paths_last_out(self):
        if len(self._path_out) > 0:
            return self.path_out[-1]

    @property
    def coords_filo(self):
        # first in and last out plus type!
        return [(inlet, {0: self.incoming, 1: self.outgoing}[nr]) for nr, inlet in
                enumerate((self.coords_first_in, self.coords_last_out)) if inlet is not None]

    # ---------------------------------------------------------------------------------------------------------------- #

    def get_inlets(self):
        if self.has_in:
            yield Inlet(coords=self.coords_in[0],
                        inlet_type=(InletTypeCodes.surface, InletTypeCodes.incoming),
                        reference=self.id,
                        frame=self.path_in[0])
            yield Inlet(coords=self.coords_in[-1],
                        inlet_type=(InletTypeCodes.internal, InletTypeCodes.incoming),
                        reference=self.id,
                        frame=self.path_in[-1])
        if self.has_out:
            yield Inlet(coords=self.coords_out[0],
                        inlet_type=(InletTypeCodes.internal, InletTypeCodes.outgoing),
                        reference=self.id,
                        frame=self.path_out[0])
            yield Inlet(coords=self.coords_out[-1],
                        inlet_type=(InletTypeCodes.surface, InletTypeCodes.outgoing),
                        reference=self.id,
                        frame=self.path_out[-1])

    def remove_inlet(self, inlet_type):
        # only surface type can be removed
        if InletTypeCodes.surface == inlet_type[0]:
            if self.has_in and InletTypeCodes.incoming == inlet_type[1]:
                self._path_object = SmartRange(self.path_in + self.path_object)
                self._path_in = SmartRange([])
                self._types_object = SmartRange(self.types_in + self.types_object)
                self._types_in = SmartRange([])
            elif self.has_out and InletTypeCodes.outgoing == inlet_type[1]:
                self._path_object = SmartRange(self.path_object + self.path_out)
                self._path_out = SmartRange([])
                self._types_object = SmartRange(self.types_object + self.types_out)
                self._types_out = SmartRange([])
        # TODO: raise exceptions or at least warnings if wrong type of inlet is used

    ####################################################################################################################
    # coords

    @property
    def coords_in(self):
        return self.single_res_selection.coords(self._path_in)

    @property
    def coords_object(self):
        return self.single_res_selection.coords(self._path_object)

    @property
    def coords_object_strict(self):
        i = (nr for nr, et in enumerate(self.etypes[1]) if et == self.etypes[1][0])
        return self.coords_object[list(i)]

    @property
    def center_of_object(self):
        return np.mean(self.coords_object_strict, 0)

    @property
    def coords_out(self):
        return self.single_res_selection.coords(self._path_out)

    @property
    def coords(self):
        return self.coords_in, self.coords_object, self.coords_out

    @property
    def coords_cont(self):
        # returns coords as one array
        return make_default_array(np.vstack([c for c in self.coords if len(c) > 0]))

    ####################################################################################################################
    # paths

    @property
    def _paths(self):
        return self._path_in, self._path_object, self._path_out

    @property
    def paths(self):
        return self.path_in, self.path_object, self.path_out

    @property
    def paths_cont(self):
        pathsc = []
        for p in self.paths:
            pathsc += p
        return pathsc

    ####################################################################################################################
    # types

    @property
    def types(self):
        # spath types
        return ([self.path_in_code] * len(self._path_in),
                [self.path_object_code] * len(self._path_object),
                [self.path_out_code] * len(self._path_out))

    @property
    def types_cont(self):
        typesc = []
        for t in self.types:
            typesc += t
        return typesc

    @property
    def gtypes(self):
        # generic types
        return self.types_in, self.types_object, self.types_out

    @property
    def gtypes_cont(self):
        gtypesc = []
        for t in self.gtypes:
            gtypesc += t
        return gtypesc

    @property
    @tupleify
    def etypes(self):
        # extended types
        for t, g in zip(self.types, self.gtypes):
            yield [t for t in zip(t, g)]

    @property
    def etypes_cont(self):
        return [t for t in zip(self.types_cont, self.gtypes_cont)]

    ####################################################################################################################

    @property
    def size(self):
        return sum(self.sizes)

    @property
    def sizes(self):
        return map(len, self._paths)

    @property
    def begins(self):
        return self.paths_cont[0]

    @property
    def ends(self):
        return self.paths_cont[-1]

    @property
    def has_in(self):
        return len(self._path_in) > 0

    @property
    def has_object(self):
        return len(self._path_object) > 0

    @property
    def has_out(self):
        return len(self._path_out) > 0

    ####################################################################################################################

    @tupleify
    def get_coords(self, smooth=None):
        # TODO: it is not used to get smooth coords but to get coords in general, conditionally smoothed
        # if smooth is not none applies smoothing
        if smooth is not None:
            for nr, coords in enumerate(self._make_smooth_coords(smooth)):
                if coords is None:
                    yield self.coords[nr]
                else:
                    yield coords
        else:
            for coords in self.coords:
                yield coords

    def _make_smooth_coords(self, smooth):
        return self.single_res_selection.coords_smooth(self._paths, smooth)

    def get_coords_cont(self, smooth=None):
        # returns coords as one array
        return make_default_array(np.vstack([c for c in self.get_coords(smooth) if len(c) > 0]))

    '''
    def apply_smoothing(self, smooth):
        # permament change!
        self.coords_in, self.coords_object, self.coords_out = self._make_smooth_coords(self.coords_cont, smooth)
    '''

    ####################################################################################################################

    def get_distance_cont(self, smooth=None, normalize=False):
        length = make_default_array(
            np.hstack((np.array([0]), np.cumsum(traces.diff(self.get_coords_cont(smooth=smooth))))))
        if normalize:
            if is_number(normalize):
                norm_factor = float(normalize)
            else:
                norm_factor = max(length)
            length = length / norm_factor
        return length

    def get_distance_rev_cont(self, *args, **kwargs):
        length = self.get_distance_cont(*args, **kwargs)
        return length[-1] - length

    @arrayify1
    def get_distance_both_cont(self, *args, **kwargs):
        length = self.get_distance_cont(*args, **kwargs)
        return (min(n, r) for n, r in zip(length, length[-1] - length))

    @arrayify1
    def get_velocity_cont(self, smooth=None, normalize=False):
        distance = self.get_distance_cont(smooth=smooth, normalize=normalize)
        return traces.derrivative(distance)

    @arrayify1
    def get_acceleration_cont(self, smooth=None, normalize=False):
        velocity = self.get_velocity_cont(smooth=smooth, normalize=normalize)
        return traces.derrivative(velocity)


class SinglePath(MacroMolPath):

    # __slots__ = "id _path_in _path_object _path_out _types_in _types_object _types_out _object_len single_res_selection".split()

    def is_single(self):
        return True

    def is_passing(self):
        return False


# TODO: there are problems with passing paths
class PassingPath(MacroMolPath):
    __slots__ = "_has_in_flag _has_out_flag".split()  # id _path_in _path_object _path_out _types_in _types_object _types_out _object_len single_res_selection".split()

    def __init__(self, path_id, paths, types):

        super(PassingPath, self).__init__(path_id, paths, types)

        self._has_in_flag = None
        self._has_out_flag = None

    def __getstate__(self):
        return self.id, self._path_in, self._path_object, self._path_out, self._types_in, self._types_object, self._types_out, self._object_len, self._has_in_flag, self._has_out_flag

    def __setstate__(self, state):
        self.id, self._path_in, self._path_object, self._path_out, self._types_in, self._types_object, self._types_out, self._object_len, self._has_in_flag, self._has_out_flag = state
        self.single_res_selection = SingleResidueSelection(self.id.id)

    def is_single(self):
        return False

    def is_passing(self):
        return True

    @property
    def has_in(self):
        return self._has_in_flag

    @has_in.setter
    def has_in(self, flag):
        self._has_in_flag = flag

    @property
    def has_out(self):
        return self._has_out_flag

    @has_out.setter
    def has_out(self, flag):
        self._has_out_flag = flag

    @property
    def coords_first_in(self):
        return self.coords_in[0]

    @property
    def paths_first_in(self):
        return self.path_in[0]

    @property
    def coords_last_out(self):
        return self.coords_in[-1]

    @property
    def paths_last_out(self):
        return self.path_in[-1]

    @property
    def types(self):
        # spath types
        return ([self.path_walk_code] * len(self._path_in),
                [],
                [])

    '''
    def get_coords(self, smooth=None):
        # TODO: it is not used to get smooth coords but to get coords in general, conditionally smoothed
        # if smooth is not none applies smoothing
        if smooth is not None:
            for nr,coords in enumerate(self._make_smooth_coords(smooth)):
                if coords is None:
                    yield self.coords[nr]
                else:
                    yield coords
        else:
            for coords in self.coords:
                yield coords

    '''

    def get_inlets(self):
        # Lets assume that if max/min_possible_frame is there is no inlet
        # how to check it?
        if self.has_in:
            yield Inlet(coords=self.coords[0][0],
                        inlet_type=(InletTypeCodes.surface, InletTypeCodes.incoming),
                        reference=self.id,
                        frame=self.paths_first_in)
        if self.has_out:
            yield Inlet(coords=self.coords[0][-1],
                        inlet_type=(InletTypeCodes.surface, InletTypeCodes.outgoing),
                        reference=self.id,
                        frame=self.paths_last_out)

    def remove_inlet(self, inlet_type):
        # only surface type can be removed
        if InletTypeCodes.surface == inlet_type[0]:
            if self.has_in and InletTypeCodes.incoming == inlet_type[1]:
                self.has_in = False
            elif self.has_out and InletTypeCodes.outgoing == inlet_type[1]:
                self.has_out = False
        # TODO: raise exceptions or at least warnings if wrong type of inlet is used


####################################################################################################################


class MasterPath(MacroMolPath):
    __slots__ = 'width_cont'.split()

    def __init__(self, sp, single_res_selection=None):
        super(MasterPath, self).__init__(sp.id, sp.paths, sp.gtypes)
        self.single_res_selection = single_res_selection
        self.width_cont = None

    def add_width(self, width):
        assert len(width) == self.size
        self.width_cont = width

    def __getstate__(self):
        return super(MasterPath, self).__getstate__() + (self.width_cont, self.single_res_selection)

    def __setstate__(self, state):
        self.id, self._path_in, self._path_object, self._path_out, self._types_in, self._types_object, self._types_out, self._object_len, self.width_cont, self.single_res_selection = state



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

# from matplotlib.colors import colorConverter

import scipy.spatial.distance as distance

from aquaduct.traj.paths import GenericPathTypeCodes as gptc
from aquaduct.traj.paths import PathTypesCodes as ptc
from aquaduct.visual.cmaps import default as default_cmap
from aquaduct.utils.helpers import zip_zip, is_number, lind

_cl2rgba = {'b': (0.0, 0.0, 1.0, 1.0),
            'c': (0.0, 0.75, 0.75, 1.0),
            'g': (0.0, 0.5, 0.0, 1.0),
            'k': (0.0, 0.0, 0.0, 1.0),
            'm': (0.75, 0, 0.75, 1.0),
            'r': (1.0, 0.0, 0.0, 1.0),
            'y': (0.75, 0.75, 0, 1.0),
            'w': (1.0, 1.0, 1.0, 1.0),
            's': (0.5, 0.5, 0.5, 1.0)}


def euclidean(A, B):
    return distance.cdist(A, B, 'euclidean')


def cityblock(A, B):
    return distance.cdist(A, B, 'cityblock')


# cc = lambda c, alpha=1.0: colorConverter.to_rgb(c)

def cc_safe(c):
    # color converter
    if c in 'rgbcmyk':
        c = _cl2rgba[c]
    c = tuple(c)
    assert len(c) in [3, 4], 'Color should be given either as one letter (rgbcmyk) or as rgb or rgba vector.'
    for e in c:
        assert is_number(e), 'Color vector has to be specified as numbers.'
        assert 0 <= e <= 1, 'Color vector elements have to be in range of 0 to 1.'
    return c[:3]


def cc(c):
    # color converter faster
    if c in 'rgbcmykws':
        c = _cl2rgba[c]
    return c[:3]


_dcc_is = ptc.path_in_code, gptc.scope_name
_dcc_cc = ptc.path_object_code, gptc.object_name
_dcc_cs = ptc.path_object_code, gptc.scope_name
_dcc_os = ptc.path_out_code, gptc.scope_name

_dcc_ws = ptc.path_walk_code, gptc.scope_name

_dcc_i = ptc.path_in_code
_dcc_c = ptc.path_object_code
_dcc_o = ptc.path_out_code
_dcc_w = ptc.path_walk_code

_default_color_codes = {_dcc_is: 'r',
                        _dcc_cc: 'g',
                        _dcc_cs: 'y',
                        _dcc_os: 'b',
                        _dcc_ws: 's',
                        _dcc_i: 'r',
                        _dcc_c: 'g',
                        _dcc_o: 'b',
                        _dcc_w: 's'}

default_color_codes = _default_color_codes


def color_codes(code, custom_codes=None):
    if custom_codes is None:
        return default_color_codes[code]
    else:
        return custom_codes[code]


def get_cmap(size):
    return [e[0][0] for e in zip_zip(default_cmap, N=size)]


class ColorMapDistMap(object):
    grey = (0.5, 0.5, 0.5, 1)

    def __init__(self):
        self.cmap = default_cmap
        self.cmap = self.__do_cadex()

    def distance(self, E1, E2):
        # E1 and E2 are RGBs matrices
        D = []
        for e1 in E1:
            DD = []
            for e2 in E2:
                DD.append(self.color_distance(e1, e2))
            D.append(DD)
        return np.array(D)

    @staticmethod
    def color_distance(e1, e2):
        # e1 and e2 are colors defs of rgb components in this order
        # Taken from http://www.compuphase.com/cmetric.htm
        # scale them to 0-255 range
        def to0255(E):
            return [int(e * 255) for e in E]

        e1 = to0255(e1)
        e2 = to0255(e2)

        rmean = (e1[0] + e2[0]) / 2
        r = e1[0] - e2[0]
        g = e1[1] - e2[1]
        b = e1[2] - e2[2]
        return np.sqrt((((512 + rmean) * r * r) >> 8) + 4 * g * g + (((767 - rmean) * b * b) >> 8))

    def __do_cadex(self):
        m = len(self.cmap)  # number of objects
        k = len(self.cmap)
        # indices
        mi = []  # model
        ti = range(m)  # test
        # FIRST OBJECT
        # get distance to mean object
        Xm = np.array(self.cmap).mean(axis=0)  # mean object
        Xm.shape = (1, Xm.shape[0])  # fix shape
        Dm = self.distance(Xm, np.array(self.cmap))
        D = self.distance(np.array(self.cmap), np.array(self.cmap))
        # min value will be the first object
        # mi.append(ti.pop(Dm.argmin()))
        mi.append(ti.pop(-1))
        if k > 1:
            # SECOND OBJECT
            Xm = np.array(self.cmap)[mi[-1]]
            Xm.shape = (1, Xm.shape[0])  # fix shape
            Dm = self.distance(Xm, np.array(self.cmap)[ti, :])
            # max value will be the first object
            mi.append(ti.pop(Dm.argmax()))
            if k > 2:
                while (len(mi) < k):
                    Dm = D[:, ti][mi, :]
                    mi.append(ti.pop(Dm.min(axis=0).argmax()))
        return lind(self.cmap, mi + ti)

    def __call__(self, node):
        if 0 < node:
            # cycle...
            return self.cmap[(node - 1) % len(self.cmap)][:3]
            # return self.cmap[int(np.round(self.cm_size * f_like(node)))][:3]
        # return grey otherwise
        return self.grey[:3]


def f_like(n):
    if n == 1:
        return 0.0
    if n == 2:
        return 0.5
    n -= 1
    order = np.floor(np.log(n) / np.log(2))
    parts = 2 ** order
    current = n - parts
    return 0.5 / parts + 1. / parts * current



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import colorsys
import itertools

import matplotlib.patches as mpatches
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import numpy as np


def color_gen():
    colors = ["#e6194b", "#3cb44b", "#ffe119", "#4363d8", "#f58231", "#911eb4", "#46f0f0",
              "#f032e6", "#bcf60c", "#fabebe", "#008080", "#e6beff", "#9a6324",
              "#800000", "#aaffc3", "#808000", "#ffd8b1", "#000075", "#808080"]
    for color in itertools.cycle(colors):
        yield color


def hex2rgb(color):
    """
    Convert HEX color to RGB format.
    :param color: String hex color.
    :return: Tuple with RGB values.
    """
    # https://stackoverflow.com/questions/29643352/converting-hex-to-rgb-value-in-python
    return tuple(int(color[i:i + 2], 16) for i in (0, 2, 4))


def polar2point(angle, r):
    """
    Transform polar coordinates to cartesian coordinates.
    :param angle: Angle.
    :param r: Radius.
    """
    rad = np.deg2rad
    return np.array([r * np.sin(rad(angle)), r * np.cos(rad(angle))])


def generate_arc(r, sa, ea, max_angle=5, reversed_=False):
    """
    Generate arc vertrices with control points for quadratic Bezier curve.

    :param r: Radius
    :param sa: Start angle.
    :param ea: End angle
    :param max_angle: Max. angle for which control point will be calculated. If > 90 curves will be significantly distorted.
    :param reversed_: If True vertices will start from ea.

    If `ea - sa > max_angle`, then `ea - sa` will be divided by `max_angle` into parts and control point will be calculated for each part.
    """
    vertices = []

    angles = []
    n = ea - sa
    while True:
        if n - max_angle >= 0:
            angles.append(max_angle)

            if n - max_angle == 0:
                break

            n -= max_angle
        else:
            angles.append(n % max_angle)
            break

    if reversed_:
        sa = -ea
        angles = reversed(angles)

    vertices.append(polar2point(np.abs(sa), r))

    for angle in angles:
        cp_angle = np.abs(sa + angle / 2.)
        # https://stackoverflow.com/questions/1734745/how-to-create-circle-with-b%C3%A9zier-curves
        cp_length = r + r * np.pi * 0.1 * pow(angle / 90., 2)

        vertices.append(polar2point(cp_angle, cp_length))
        vertices.append(polar2point(np.abs(sa + angle), r))

        sa += angle

    return vertices


class Node(mpatches.PathPatch):
    def __init__(self, r, sa, ea, color):
        """
        Represent data on Chord circle.

        :param r: Radius.
        :param sa: Start angle.
        :param ea: End angle.
        :param color: Node color in HEX or matplotlib tuple format.
        """
        self.sa = sa
        self.ea = ea
        self.color = color

        # Remembers when last link was added
        self._link_angle = sa

        vertices = []
        codes = []

        # # Outer arc
        outer_arc = generate_arc(r, sa, ea)
        vertices.extend(outer_arc)

        codes.append(mpath.Path.MOVETO)
        codes.extend([mpath.Path.CURVE3] * (len(outer_arc) - 1))

        # Inner arc
        inner_arc = generate_arc(0.8 * r, sa, ea, reversed_=True)
        vertices.extend(inner_arc)

        codes.append(mpath.Path.LINETO)
        codes.extend([mpath.Path.CURVE3] * (len(inner_arc) - 1))

        # Closing poly
        vertices.append((0, 0))
        codes.append(mpath.Path.CLOSEPOLY)

        path = mpath.Path(vertices, codes)

        super(Node, self).__init__(path, facecolor=color, linewidth=0)

    def reserve_arc(self, angle):
        # Only if scaling is an optional
        # if self._link_angle + angle - self.sa > self.ea - self.sa:
        #     raise RuntimeError("Too much data for node. Use scaling option.")

        self._link_angle += angle

    def get_arc_offset(self):
        return self._link_angle


class Link(mpatches.PathPatch):
    def __init__(self, r, sa0, sa1, ea0, ea1, color):
        """
        Represent connection between two nodes.

        :param r: Radius.
        :param sa0: Source start angle.
        :param sa1: Source end angle.
        :param ea0: Destination start angle.
        :param ea1: Destination end angle.
        :param color: Link color in HEX or matplotlib tuple format.
        """
        vertices = []
        codes = []

        # Source arc
        source_arc = generate_arc(r, sa0, sa1)
        vertices.extend(source_arc)

        codes.append(mpath.Path.MOVETO)
        codes.extend([mpath.Path.CURVE3] * (len(source_arc) - 1))

        # Connection line
        vertices.append(polar2point(sa1, r))
        codes.append(mpath.Path.LINETO)

        vertices.append((0, 0))
        codes.append(mpath.Path.CURVE3)

        vertices.append(polar2point(ea0, r))
        codes.append(mpath.Path.CURVE3)

        # Dest arc
        dest_arc = generate_arc(r, ea0, ea1)
        vertices.extend(dest_arc)

        codes.extend([mpath.Path.CURVE3] * len(dest_arc))

        # Connection line
        vertices.append(polar2point(ea1, r))
        codes.append(mpath.Path.LINETO)

        vertices.append((0, 0))
        codes.append(mpath.Path.CURVE3)

        vertices.append(polar2point(sa0, r))
        codes.append(mpath.Path.CURVE3)

        path = mpath.Path(vertices, codes)
        super(Link, self).__init__(path, facecolor=color, linewidth=0, alpha=0.6)


class Arrow(mpatches.PathPatch):
    def __init__(self, r, sa, ea, color, max_angle=45):
        """
        Arrow patch for links.

        :param r: Radius.
        :param sa: Start angle.
        :param ea: End angle.
        :param color: Arrow color in HEX or matplotlib tuple format.
        :param max_angle:
        """
        vertices = []
        codes = []

        c = sa + (ea - sa) / 2

        if ea - sa < max_angle:
            max_angle = ea - sa

        arc = generate_arc(r, c - max_angle / 2, c + max_angle / 2)
        vertices.extend(arc)

        codes.append(mpath.Path.MOVETO)
        codes.extend([mpath.Path.CURVE3] * (len(arc) - 1))

        vertices.append(polar2point(sa + (ea - sa) / 2, 0.9 * r - 2))
        codes.append(mpath.Path.LINETO)

        vertices.append((0, 0))
        codes.append(mpath.Path.CLOSEPOLY)

        path = mpath.Path(vertices, codes)
        super(Arrow, self).__init__(path, facecolor=color, linewidth=0)


class Chord(object):
    def __init__(self, ax, r, nodes_sizes, links, labels, colors=[]):
        """
        
        :param ax:
        :param r:
        :param nodes_sizes:
        :param links:
        :param labels:
        :param colors:
        """
        self.nodes = []

        if colors:
            self.colors = colors
        else:
            cg = color_gen()
            self.colors = [next(cg) for _ in range(0, len(nodes_sizes))]

        sizes_sum = sum(nodes_sizes)

        # Nodes
        legend_node_index = []  # Keeps indexes of nodes which will be in legend
        sa = 0
        for i, size in enumerate(nodes_sizes):
            ea = sa + size * 360. / sizes_sum

            if ea - sa >= 10:
                angle = sa + (ea - sa) / 2
                pos = polar2point(angle, 1.05 * r)
                ax.text(pos[0], pos[1],
                        "{} ({:.2f}%)".format(labels[i], 100.0 * size / sizes_sum),
                        verticalalignment="center",
                        horizontalalignment="center",
                        fontsize=10,
                        rotation=360 - angle)
            else:
                legend_node_index.append(i)

            node = Node(r, sa, ea, self.colors[i])
            ax.add_patch(node)

            self.nodes.append(node)

            sa = ea

        # Create legend
        ax.legend([self.nodes[i] for i in legend_node_index],
                  ["{} ({:.2f}%)".format(labels[i], 100.0 * nodes_sizes[i] / sizes_sum) for i in legend_node_index],
                  bbox_to_anchor=(1.02, 1), loc="upper left")

        # Scales
        links_sum = [0.] * len(nodes_sizes)
        for link in links:
            links_sum[link["source"]] += link["value"]
            links_sum[link["dest"]] += link["value"]

        scales = []
        for i, (size, link_sum) in enumerate(zip(nodes_sizes, links_sum)):
            if size < link_sum:
                scales.append(0.9999 * size / link_sum)  # Scale is reduced to 99% due to float precision
            else:
                scales.append(1.)

        # Links
        for link in links:
            source_node = self.nodes[link["source"]]
            dest_node = self.nodes[link["dest"]]
            value = link["value"]
            sa0 = source_node.get_arc_offset()
            slink_arc = scales[link["source"]] * value * 360. / sizes_sum
            sa1 = sa0 + slink_arc

            source_node.reserve_arc(slink_arc)

            ea0 = dest_node.get_arc_offset()
            elink_arc = scales[link["dest"]] * value * 360. / sizes_sum
            ea1 = ea0 + elink_arc

            dest_node.reserve_arc(elink_arc)

            l = Link(0.8 * r, sa0, sa1, ea0, ea1, source_node.color)
            a = Arrow(0.81 * r, sa0, sa1, source_node.color)

            ax.add_patch(l)
            ax.add_patch(a)

            if sa1 - sa0 > 6:
                # Complimentary color
                rgb_color = hex2rgb(source_node.color.lstrip("#")) if isinstance(source_node.color,
                                                                                 str) else source_node.color
                hsl_color = list(colorsys.rgb_to_hls(*rgb_color))

                if not hsl_color[0]:
                    if 138 >= hsl_color[1] >= 118:  # For gray color automatically set white
                        hsl_color[1] = 255
                    else:
                        hsl_color[1] = (hsl_color[1] - 255) * -1.0
                else:
                    hsl_color[0] += 0.5

                complimentary_color = [c / 255 for c in colorsys.hls_to_rgb(*hsl_color)]

                # Arrow text
                angle = sa0 + (sa1 - sa0) / 2
                pos = polar2point(sa0 + (sa1 - sa0) / 2, 0.86 * r)
                ax.text(pos[0], pos[1],
                        link["value"],
                        verticalalignment="center",
                        horizontalalignment="center",
                        fontsize=9,
                        rotation=180 - angle,
                        color="white")


if __name__ == "__main__":
    fig, ax = plt.subplots(subplot_kw={"aspect": 1})
    ax.set_axis_off()
    ax.set_xlim(-110, 110)
    ax.set_ylim(-110, 110)

    labels = ["Data #1", "Data #2", "Data #3", "Data #4"]

    sizes = [5000, 6000, 7000]

    links = [dict(source=1, dest=0, value=2000),
             dict(source=1, dest=0, value=2000),
             dict(source=2, dest=0, value=4600),
             dict(source=0, dest=0, value=1000)]

    Chord(ax, 100, sizes, links, labels)

    fig.savefig("chord.png", format="png", dpi=2 ** 7)

    plt.show()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2018-2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import find_packages, setup

version = __import__('aquaduct').version_nice()  # version str

install_requires = ['numpy>=1.10.0',  # this is required by MDA
                    'scipy>=0.17.1',
                    'scikit-learn>=0.16.0',
                    'MDAnalysis[amber]==0.16.2',
                    'joblib>=0.13'
                    ]


def install_requires_nice(level=0):
    import re
    for ir in install_requires:
        print (" " * 4 * level) + "* " + " ".join(re.split('(>=|==|<=|>|<|=)', ir))


author = __import__('aquaduct').__author__  # version str

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(name='aquaduct',
      version=version,
      description='Tracing molecules in MD simulation',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://aquaduct.pl',
      project_urls={
          'Documentation': 'http://aquaduct.pl/documentation/',
          'Source': 'https://github.com/tljm/aqua-duct/',
          'Tracker': 'https://github.com/tljm/aqua-duct/issues/',
      },
      author=author,
      author_email='info@aquaduct.pl',
      license='GNU GPL v3',
      keywords='molecular-dynamics solvent',
      packages=find_packages(include=['aquaduct*']),
      python_requires='>=2.7, <3',
      install_requires=install_requires,
      extras_require={'full_pymol': ["pymol>=1.4"],
                      'graphs': ['matplotlib'],
                      },
      scripts=['apps/valve.py', 'apps/valve_run',
               'apps/valveconfig.py', 'apps/valveconfig_run',
               'apps/portal.py', 'apps/portal_run',
               'apps/pond.py', 'apps/pond_run',
               'apps/kraken.py', 'apps/kraken_run',
               'apps/hs_resize.py', ],
      provides=['aquaduct'],
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'Operating System :: POSIX',
                   'Programming Language :: Python :: 2.7',
                   ],
      include_package_data=True
      )



# -*- coding: utf-8 -*-

# This program is distributed in the hope that it will be useful,
# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
This module is meant to read configuration file and init some global options.
It cannot import sandwich. Directly or indirectly.
'''

from aquaduct import logger, greetings as greetings_aquaduct, version_nice as aquaduct_version_nice

import ConfigParser
import os
from collections import OrderedDict, namedtuple
from keyword import iskeyword

from aquaduct.apps.data import GCS, load_cric
from aquaduct.geom.cluster_available_methods import get_required_params, \
    AVAILABLE_METHODS as available_clustering_methods
from aquaduct.utils import clui
from aquaduct.utils.helpers import Auto


class ConfigSpecialNames(object):
    special_names_dict = {'none': None,
                          'null': None,
                          'true': True,
                          'false': False,
                          'yes': True,
                          'no': False,
                          'auto': Auto}

    def special_name(self, name):
        if isinstance(name, (str, unicode)):
            if name.lower() in self.special_names_dict:
                return self.special_names_dict[name.lower()]
        return name


class ValveConfig(ConfigSpecialNames):
    def __init__(self):
        self.config = self.get_default_config()
        self.config_filename = ''

    def __make_options_nt(self, input_options):
        # options = {opt: self.special_name(input_options[opt]) for opt in input_options}
        options = list()
        for opt in input_options:
            if iskeyword(opt):
                logger.warning('Invalid keyword <%s> in config file skipped. Check configuration file.' % opt)
                continue
            if opt.replace('_', '').isalnum():
                options.append((opt, self.special_name(input_options[opt])))
            else:
                logger.debug('Keyword <%s> in config file skipped.' % opt)
        options = OrderedDict(options)
        options_nt = namedtuple('Options', options.keys())
        return options_nt(**options)

    @staticmethod
    def common_config_names():
        # execute - what to do: skip, run
        # load - load previous results form file name
        # save - save results to file name
        return 'execute dump'.split()

    @staticmethod
    def common_traj_data_config_names():
        # scope - scope definition
        # scope_convexhull - take convex hull of scope, true of false
        # object - object definition
        return 'scope scope_convexhull scope_everyframe scope_convexhull_inflate object'.split()

    @staticmethod
    def global_name():
        return 'global'

    @staticmethod
    def cluster_name():
        return 'clustering'

    @staticmethod
    def recluster_name():
        return 'reclustering'

    @staticmethod
    def recursive_clustering_name():
        return 'recursive_clustering'

    @staticmethod
    def recursive_threshold_name():
        return 'recursive_threshold'

    @staticmethod
    def smooth_name():
        return 'smooth'

    def stage_names(self, nr=None):
        if nr is None:
            return [self.stage_names(nr) for nr in range(5)]
        else:
            if nr == 0:
                return 'traceable_residues'
            elif nr == 1:
                return 'raw_paths'
            elif nr == 2:
                return 'separate_paths'
            elif nr == 3:
                return 'inlets_clustering'
            elif nr == 4:
                return 'analysis'
            elif nr == 5:
                return 'visualize'
        raise NotImplementedError('Stage %r is not implemented.' % nr)

    def get_common_traj_data(self, stage):
        assert isinstance(stage, int)
        # options = {name: None for name in self.common_traj_data_config_names()}
        options = OrderedDict(((name, None) for name in self.common_traj_data_config_names()))
        for nr in range(stage + 1)[::-1]:
            section = self.stage_names(nr)
            for name in self.common_traj_data_config_names():
                if self.config.has_option(section, name):
                    value = self.special_name(self.config.get(section, name))
                    if (value is not None) and (options[name] is None):
                        options.update({name: value})
        return self.__make_options_nt(options)

    def get_global_options(self):
        section = self.global_name()
        names = self.config.options(section)
        # options = {name: self.config.get(section, name) for name in names}
        options = OrderedDict(((name, self.config.get(section, name)) for name in names))
        return self.__make_options_nt(options)

    def get_stage_options(self, stage):
        assert isinstance(stage, int)
        stage_name = self.stage_names(stage)
        names = self.config.options(stage_name)
        # options = {name: self.config.get(stage_name, name) for name in names}
        options = OrderedDict(((name, self.config.get(stage_name, name)) for name in names))
        if stage in [0, 1]:
            options.update(self.get_common_traj_data(stage)._asdict())
        return self.__make_options_nt(options)

    def get_cluster_options(self, section_name=None):
        if section_name is None:
            section = self.cluster_name()
        else:
            section = section_name
        names = self.config.options(section)
        # options = {name: self.config.get(section, name) for name in names}
        options = OrderedDict(((name, self.config.get(section, name)) for name in names))
        # special recursive_clustering option
        if self.recursive_clustering_name() not in options:
            options.update({self.recursive_clustering_name(): None})
        if self.recursive_threshold_name() not in options:
            options.update({self.recursive_threshold_name(): None})
        return self.__make_options_nt(options)

    def get_recluster_options(self):
        return self.get_cluster_options(section_name=self.recluster_name())

    def get_smooth_options(self):
        section = self.smooth_name()
        names = self.config.options(section)
        # options = {name: self.config.get(section, name) for name in names}
        options = OrderedDict(((name, self.config.get(section, name)) for name in names))
        return self.__make_options_nt(options)

    def get_default_config(self):
        # snr = 0 # stage number

        config = ConfigParser.RawConfigParser()

        def common(section):
            for setting in self.common_config_names():
                value = None
                if setting == 'execute':
                    value = 'runonce'
                elif setting == 'dump':
                    if GCS.netcdf:
                        value = '%d_%s_data.nc' % (snr + 1, section)
                    else:
                        value = '%d_%s_data.dump' % (snr + 1, section)
                if value is None:
                    config.set(section, setting)
                else:
                    config.set(section, setting, value=value)

        def common_traj_data(section, commented=False):
            for setting in self.common_traj_data_config_names():
                if commented:
                    setting = '#' + setting
                config.set(section, setting, 'None')

        ################
        # global settings
        section = self.global_name()
        config.add_section(section)

        # top - top file name
        # nc - netcdf file name
        config.set(section, 'top')  # topology
        config.set(section, 'trj')  # trajectory

        config.set(section, 'twoway', 'True')
        config.set(section, 'sps', 'True')

        config.set(section, 'sandwich', 'False')
        config.set(section, 'waterfall', 'False')
        config.set(section, 'max_frame', 'None')
        config.set(section, 'min_frame', 'None')
        config.set(section, 'step_frame', 'None')

        config.set(section, 'cache_dir', 'None')
        config.set(section, 'cache_mem', 'False')

        ################
        snr = 0  # stage number
        # stage I
        # find traceable residues
        section = self.stage_names(snr)
        config.add_section(section)

        common(section)
        common_traj_data(section)
        config.set(section, 'scope_everyframe', 'False')
        config.set(section, 'scope_convexhull_inflate', 'None')

        config.set(section, 'add_passing', 'None')

        ################
        snr += 1
        # stage II
        # find raw paths
        section = self.stage_names(snr)
        config.add_section(section)

        common(section)
        common_traj_data(section)

        config.set(section, 'clear_in_object_info', 'False')
        config.set(section, 'discard_singletons', 1)
        config.set(section, 'discard_empty_paths', 'True')

        ################
        snr += 1
        # stage III
        # create separate frames
        section = self.stage_names(snr)
        config.add_section(section)

        common(section)

        config.set(section, 'allow_passing_paths', 'False')
        config.set(section, 'auto_barber_tovdw', 'True')
        config.set(section, 'auto_barber_maxcut', '2.8')
        config.set(section, 'auto_barber_mincut', 'None')
        config.set(section, 'auto_barber_maxcut_level', 'True')
        config.set(section, 'auto_barber_mincut_level', 'True')
        config.set(section, 'auto_barber', 'False')
        config.set(section, 'separate_barber', 'True')
        config.set(section, 'discard_empty_paths', 'True')
        config.set(section, 'sort_by_id', 'True')
        config.set(section, 'apply_smoothing', 'False')
        config.set(section, 'apply_soft_smoothing', 'True')
        config.set(section, 'remove_unused_parts', 'True')

        config.set(section, 'calculate_coo', 'False')

        config.set(section, 'discard_short_paths', '20')
        config.set(section, 'discard_short_object', '2.0')
        config.set(section, 'discard_short_logic', 'or')

        ################
        snr += 1
        # stage IV
        # inlets clusterisation
        section = self.stage_names(snr)
        config.add_section(section)

        common(section)

        config.set(section, 'max_level', '0')
        config.set(section, 'recluster_outliers', 'False')
        config.set(section, 'detect_outliers', 'False')
        config.set(section, 'singletons_outliers', 'False')
        config.set(section, 'create_master_paths', 'False')
        config.set(section, 'exclude_passing_in_clustering', 'True')
        config.set(section, 'add_passing_to_clusters', 'None')
        config.set(section, 'renumber_clusters', 'False')
        config.set(section, 'join_clusters', 'None')
        config.set(section, 'master_paths_amount', 'None')
        config.set(section, 'separate_master', 'False')
        config.set(section, 'separate_master_all', 'True')
        config.set(section, 'inlets_center', 'cos')
        config.set(section, 'remove_inlets', 'None')
        config.set(section, 'clustering_order', 'old-school')

        ################
        # smooth
        section = self.smooth_name()
        config.add_section(section)
        config.set(section, 'method', 'window')

        ################
        # clustering
        section = self.cluster_name()
        config.add_section(section)
        config.set(section, 'method', 'barber')
        config.set(section, self.recursive_clustering_name(), self.cluster_name())
        config.set(section, self.recursive_threshold_name(), 'False')

        ################
        # reclustering
        section = self.recluster_name()
        config.add_section(section)
        config.set(section, 'method', 'dbscan')
        config.set(section, self.recursive_clustering_name(), 'False')
        config.set(section, self.recursive_threshold_name(), 'False')

        ################
        snr += 1
        # stage V
        # analysis
        section = self.stage_names(snr)
        config.add_section(section)
        common(section)
        config.remove_option(section, 'dump')
        config.set(section, 'save', '%d_%s_results.txt' % (snr + 1, section))

        config.set(section, 'calculate_scope_object_size', 'False')
        config.set(section, 'scope_chull', 'None')
        config.set(section, 'scope_chull_inflate', 'None')
        config.set(section, 'object_chull', 'None')

        config.set(section, 'cluster_area', 'True')
        config.set(section, 'cluster_area_precision', '20')
        config.set(section, 'cluster_area_expand', '2')

        config.set(section, 'dump_config', 'True')

        ################
        snr += 1
        # stage VI
        # visualize
        section = self.stage_names(snr)
        config.add_section(section)
        common(section)
        config.remove_option(section, 'dump')
        config.set(section, 'save', '%d_%s_results.py' % (snr + 1, section))

        config.set(section, 'simply_smooths', 'RecursiveVector')
        config.set(section, 'split_by_type', 'False')
        config.set(section, 'retain_all_types', 'False')

        # visualize spaths, all paths in one object
        config.set(section, 'all_paths_raw', 'False')
        config.set(section, 'all_paths_smooth', 'False')
        config.set(section, 'all_paths_split', 'False')  # split by in obj out
        config.set(section, 'all_paths_raw_io', 'False')
        config.set(section, 'all_paths_smooth_io', 'False')
        config.set(section, 'all_paths_amount', 'None')

        # visualize spaths, separate objects
        config.set(section, 'paths_raw', 'False')
        config.set(section, 'paths_smooth', 'False')
        config.set(section, 'paths_states', 'False')
        config.set(section, 'paths_raw_io', 'False')
        config.set(section, 'paths_smooth_io', 'False')

        config.set(section, 'ctypes_raw', 'False')
        config.set(section, 'ctypes_smooth', 'False')
        config.set(section, 'ctypes_amount', 'None')

        # visualize clusters
        config.set(section, 'inlets_clusters', 'False')
        config.set(section, 'inlets_clusters_amount', 'None')

        config.set(section, 'cluster_area', 'False')
        config.set(section, 'cluster_area_precision', '20')
        config.set(section, 'cluster_area_expand', '2')

        # show protein
        config.set(section, 'show_molecule', 'None')
        config.set(section, 'show_molecule_frames', '0')
        config.set(section, 'show_scope_chull', 'None')
        config.set(section, 'show_scope_chull_inflate', 'None')
        config.set(section, 'show_scope_chull_frames', '0')
        config.set(section, 'show_object_chull', 'None')
        config.set(section, 'show_object_chull_frames', '0')

        return config

    def load_config(self, filename):
        self.config.read(filename)
        self.config_filename = filename

    def save_config_stream(self, fs):
        self.config.write(fs)

    def save_config(self, filename):
        with open(filename, 'w') as fs:
            self.save_config_stream(fs)

    def get_general_comment(self, section):
        if section == self.global_name():
            return ['Global settings.']
        if section == self.cluster_name():
            out = ['Possible clustering methods:']
            # get default clustering method
            defmet = self.config.get(section, 'method')
            for method in available_clustering_methods:
                if method == defmet:
                    continue
                out.append('method = %s' % method)
                params = get_required_params(method)
                if params:
                    sss = ''
                    if len(params) > 1:
                        sss = 's'
                    out.append('Required parameter%s for %s method:' % (sss, method))
                    for param in params:
                        out.append('%s = None' % param)
                    # out.append('')
            return out

    def dump_config(self, dump_template=False):
        skip_list = '''simply_smooths'''.split()  # these options are very optional!
        concise = True
        if dump_template:
            concise = False

        output = []

        options = [self.get_global_options()] + \
                  [self.get_stage_options(stage) for stage in range(6)] + \
                  [self.get_cluster_options(), self.get_recluster_options(),
                   self.get_smooth_options()]
        names = [self.global_name()] + \
                [self.stage_names(stage) for stage in range(6)] + \
                [self.cluster_name(), self.recluster_name(),
                 self.smooth_name()]

        def value2str(val):
            if val is Auto:
                val = "Auto"
            else:
                val = str(val)
            return val

        # this loop ensures it is returned in particular order
        for opts, name in zip(options, names):
            output.append('[%s]' % name)  # section name
            # general comments
            if dump_template:
                comment = self.get_general_comment(name)
                if comment:
                    output.extend(['# %s' % line for line in comment])
            for key, value in opts._asdict().iteritems():  # loop over options
                if key in skip_list:
                    continue
                # comment scope etc. in stage II if dump_template
                if dump_template and name == self.stage_names(1):
                    if key in self.common_traj_data_config_names():
                        key = '#' + key
                output.append('%s = %s' % (key, value2str(value)))
            if not concise:
                output.append('')

        # is something missing? another loop over all additional sections
        for miss in self.config.sections():
            if miss in names:
                continue  # skip if it was already dumped in the loop above
            output.append('[%s]' % miss)
            for key in self.config.options(miss):  # loop over options in section miss
                if key in skip_list:
                    continue
                value = self.config.get(miss, key)
                output.append('%s = %s' % (key, value2str(value)))
            if not concise:
                output.append('')
        while not len(output[-1].strip()):
            output.pop()
        return output


def valve_load_config(filename, config):
    assert filename is not None, "No config file provided."
    assert os.path.isfile(filename), "Config file %s does not exist." % filename
    with clui.fbm('Load configuration file'):
        config.load_config(filename)
        goptions = config.get_global_options()

        # cache dir & netcdf
        GCS.cachedir = goptions.cache_dir
        GCS.cachemem = goptions.cache_mem
        # GCS.netcdf = args.use_netcdf or args.use_netcdf4
        # GCS.nc4 = args.use_netcdf4
        load_cric()

        # single precision storage
        if goptions.sps:
            logger.info('Single precision data storage activated.')
            from aquaduct.utils.maths import defaults
            from numpy import float32, int32

            defaults.float_default = float32
            defaults.int_default = int32


def sep():
    return clui.gsep(sep='-', times=48)


def asep():
    return clui.gsep(sep='=', times=72)


def valve_begin():
    clui.message(greetings_aquaduct())  # nice greetings
    clui.message('Aqua-Duct version %s' % aquaduct_version_nice())
    # clui.message('Valve driver version %s' % version_nice())
    clui.message(sep())


def valve_end():
    clui.message(sep())
    clui.message('Let the Valve be always open!')
    clui.message('Goodbye!')



# -*- coding: utf-8 -*-

# This program is distributed in the hope that it will be useful,
# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps

import numpy as np
from scipy.stats import ttest_ind
from aquaduct.traj.inlets import InletTypeCodes
from aquaduct.apps.valve.spath import spath_lenght_total_info, spath_frames_info
from aquaduct.geom import hdr


def cluster_id_header():
    return ['Cluster'], ['%7d']


def add_cluster_id_head(gen):
    sph, splt = cluster_id_header()

    @wraps(gen)
    def patched(*args, **kwargs):
        h, lt = gen(*args, **kwargs)
        return sph + h, splt + lt

    return patched


def add_cluster_id(gen):
    @wraps(gen)
    def patched(cluster, something, add_id=True, *args, **kwargs):
        line = gen(cluster, something, *args, **kwargs)
        if add_id:
            line = [int(cluster)] + line
        return line

    return patched


@add_cluster_id_head
def clusters_inlets_header():
    header = 'Size SInp IInp IOut SOut'.split()
    header = 'Size INCOMING OUTGOING'.split()
    line_template = ['%7d'] + ['%8d'] * (len(header) - 1)
    return header, line_template


@add_cluster_id
def clusters_inlets(cluster, inlets):
    line = [inlets.size]
    line.append(inlets.lim2types([InletTypeCodes.surface_incoming]).size)
    # line.append(inlets.lim2types([InletTypeCodes.internal_incoming]).size)
    # line.append(inlets.lim2types([InletTypeCodes.internal_outgoing]).size)
    line.append(inlets.lim2types([InletTypeCodes.surface_outgoing]).size)
    return line


@add_cluster_id_head
def clusters_area_header():
    header = ('D' + ' D'.join(map(str, range(100, 85, -5) + range(80, 40, -10)))).split()
    line_template = ['%8.2f'] * len(header)
    return header, line_template


@add_cluster_id
def clusters_area(cluster, inlets, points=10, expand_by=1):
    # HDR
    h = hdr.HDR(np.array(inlets.coords), points=points, expand_by=expand_by, center_of_system=inlets.center_of_system)
    line = list()
    for fraction in range(100, 85, -5) + range(80, 40, -10):
        line.append(h.area(fraction=fraction / 100.))
    return line


@add_cluster_id_head
def clusters_stats_prob_header():
    header = 'IN-OUT diff N IN-OUT_prob diff_prob N_prob'.split()
    line_template = ['%8d'] * (len(header) / 2) + ['%12.2f'] * (len(header) / 2)
    # header += 'IN_len OUT_len Both_len'.split()
    # line_template += ['%9.1f'] * 3
    return header, line_template


@add_cluster_id
def clusters_stats_prob(cluster, sp_ct):
    # calculates probabilities of some events for cluster
    # X:X transition - io
    # X:? and ?:X transition - d
    # X:N and N:X transition - N
    line = []
    io, d, N = 0, 0, 0
    in_len, out_len, tot_len = 0., 0., 0.
    in_n, out_n, tot_n = 0, 0, 0
    for sp, ct in sp_ct:
        ct = ct.generic.clusters
        assert cluster in ct
        if cluster == ct[0] and cluster == ct[1]:
            io += 1
        elif None in ct:
            N += 1
        else:
            d += 1
    line += [io, d, N]
    summa = float(sum([io, d, N]))
    line += map(lambda x: x / summa if summa else float('nan'), [io, d, N])
    return line


@add_cluster_id_head
def clusters_stats_len_header():
    header = 'X->Obj Obj->X p-value X->ObjMin X->ObjMinID Obj->XMin Obj->XMinID'.split()
    line_template = (['%9.1f'] * 2) + ['%9.4f'] + ['%9.1f', '%11s'] * 2
    return header, line_template


@add_cluster_id
def clusters_stats_len(cluster, sp_ct):
    line = []
    in_len, out_len = [], []
    in_len_min, out_len_min = float('inf'), float('inf')
    in_len_min_id, out_len_min_id = None, None
    for sp, ct in sp_ct:
        ct = ct.generic.clusters
        assert cluster in ct
        lens = spath_lenght_total_info(sp, add_id=False, total=False, totalonly=False)
        # tot,in,obj,out
        if cluster == ct[0]:
            if not np.isnan(lens[0]):
                in_len.append(lens[0])
                if lens[0] < in_len_min:
                    in_len_min = lens[0]
                    in_len_min_id = str(sp.id)
        if cluster == ct[1]:
            if not np.isnan(lens[-1]):
                out_len.append(lens[-1])
                if lens[-1] < out_len_min:
                    out_len_min = lens[-1]
                    out_len_min_id = str(sp.id)

    if len(in_len):
        line.append(np.mean(in_len))
    else:
        line.append(float('nan'))
    if len(out_len):
        line.append(np.mean(out_len))
    else:
        line.append(float('nan'))
    if len(in_len) > 1 and len(out_len) > 1:
        line.append(ttest_ind(in_len, out_len)[-1])  # this is supposed to return p-value
    else:
        line.append(float('nan'))

    line += [in_len_min, in_len_min_id, out_len_min, out_len_min_id]

    return line


@add_cluster_id_head
def clusters_stats_steps_header():
    header = 'X->Obj Obj->X p-value X->ObjMin X->ObjMinID Obj->XMin Obj->XMinID'.split()
    line_template = (['%9.1f'] * 2) + ['%9.4f'] + ['%9.1f', '%11s'] * 2
    return header, line_template


@add_cluster_id
def clusters_stats_steps(cluster, sp_ct):
    line = []
    in_len, out_len = [], []
    in_len_min, out_len_min = float('inf'), float('inf')
    in_len_min_id, out_len_min_id = None, None
    for sp, ct in sp_ct:
        ct = ct.generic.clusters
        assert cluster in ct
        lens = spath_frames_info(sp, add_id=False, total=False)
        # tot,in,obj,out
        if cluster == ct[0]:
            if not np.isnan(lens[0]):
                in_len.append(lens[0])
                if lens[0] < in_len_min:
                    in_len_min = lens[0]
                    in_len_min_id = str(sp.id)
        if cluster == ct[1]:
            if not np.isnan(lens[-1]):
                out_len.append(lens[-1])
                if lens[-1] < out_len_min:
                    out_len_min = lens[-1]
                    out_len_min_id = str(sp.id)

    if len(in_len):
        line.append(np.mean(in_len))
    else:
        line.append(float('nan'))
    if len(out_len):
        line.append(np.mean(out_len))
    else:
        line.append(float('nan'))
    if len(in_len) > 1 and len(out_len) > 1:
        line.append(ttest_ind(in_len, out_len)[-1])  # this is supposed to return p-value
    else:
        line.append(float('nan'))

    line += [in_len_min, in_len_min_id, out_len_min, out_len_min_id]

    return line
